
#include <stdio.h>
#include <string.h>
#include <sys/timeb.h>

#include "template.h"

class Instance {
protected:
  struct Vec { int vaha, cena; };
  int ID,n,M;
  CArray<Vec> veci;
public:
  Instance() { }
  Instance(const char *line) {
    int read;
    if (sscanf(line,"%d %d %d%n",&ID,&n,&M,&read)<3) throw "Chyba pri cteni souboru";
    line += read;
    veci.Init(n);
    for (int i=0; i<n; i++) {
      if (sscanf(line,"%d %d%n",&veci[i].vaha,&veci[i].cena,&read)<2) throw "Chyba pri cteni souboru";
      line += read;
    }
  }
  int Pocet() {
    return n;
  }
  int Limit() {
    return M;
  }
  int Id() {
    return ID;
  }
  int Vaha(int i) {
    return veci[i].vaha;
  }
  int Cena(int i) {
    return veci[i].cena;
  }
};

class Konfigurace {
protected:
  Instance instance;
  CArray<bool> veci;
  int cena, vaha, vybrano;
public:
  void Init() {
    for (int i=0; i<instance.Pocet(); i++) veci[i] = false;
    cena = vaha = vybrano = 0;
  }
  Konfigurace() { }
  Konfigurace(Instance instance) {
    this->instance = instance;
    veci.Init(instance.Pocet());
    Init();
  }
  bool Vybrano(int vec) {
    return veci[vec];
  }
  void Pridej(int vec) {
    if (Vybrano(vec)) throw "Opakovane pridani";
    veci[vec] = true;
    cena += instance.Cena(vec);
    vaha += instance.Vaha(vec);
    vybrano++;
  }
  void Uber(int vec) {
    if (!Vybrano(vec)) throw "Opakovane ubrani";
    veci[vec] = false;
    cena -= instance.Cena(vec);
    vaha -= instance.Vaha(vec);
    vybrano--;
  }
  int Cena() {
    return cena;
  }
  int Vaha() {
    return vaha;
  }
  int PocetVybranych() {
    return vybrano;
  }
  bool Pretizen() {
    return (Vaha()>instance.Limit());
  }
  void Vypis() {
    for (int i=0; i<instance.Pocet(); i++) {
      printf(Vybrano(i) ? "1" : "0");
    }
    printf("\n");
  }
};

class Algoritmus {
protected:
  Instance instance;
  Konfigurace nejlepsi_konfigurace;
  timeb start, stop;
protected:
  Algoritmus(Instance instance) : nejlepsi_konfigurace(instance) {
    this->instance = instance;
  }
  void StartTime() {
    ftime(&start);
  }
  void StopTime() {
    ftime(&stop);
  }
public:
  void VypisReseni(FILE *reseni) {
    fprintf(reseni,"%d %d %d",instance.Id(),instance.Pocet(),nejlepsi_konfigurace.Cena());
    for (int i=0; i<instance.Pocet(); i++) {
      fprintf(reseni," %c",nejlepsi_konfigurace.Vybrano(i) ? '1' : '0');
    }
    fprintf(reseni,"\n");
    fflush(reseni);
  }
  void VypisReport(FILE *report) {
    int ms = 1000*(stop.time-start.time) + stop.millitm - start.millitm;
    fprintf(report,"%d %d %d\n",instance.Id(),instance.Pocet(),ms);
    fflush(report);
  }
};

class HrubaSila : public Algoritmus {
protected:
  Konfigurace konfigurace;
public:
  HrubaSila(Instance instance) : Algoritmus(instance), konfigurace(instance) { }
  void Test() {
    if (konfigurace.Cena()>nejlepsi_konfigurace.Cena() && !konfigurace.Pretizen()) {
      nejlepsi_konfigurace = konfigurace;
    }
  }
  void Dalsi(int vec) {
    if (vec>=instance.Pocet() || konfigurace.Pretizen()) return;
    konfigurace.Pridej(vec);
    Test();
    Dalsi(vec+1);
    konfigurace.Uber(vec);
    Test();
    Dalsi(vec+1);
  }
  void Start() {
    StartTime();
    konfigurace.Init();
    nejlepsi_konfigurace.Init();
    Dalsi(0);
    StopTime();
  }
};

class HeuristikaCenaVaha : public Algoritmus {
protected:
  Konfigurace konfigurace;
public:
  HeuristikaCenaVaha(Instance instance) : Algoritmus(instance), konfigurace(instance) { }
  int Najdi() {
    int nejlepsi_i = 0;
    float nejlepsi_pomer = 0;
    for (int i=0; i<instance.Pocet(); i++)
      if (!konfigurace.Vybrano(i)) {
        float pomer = (float)instance.Cena(i)/instance.Vaha(i);
        if (pomer>nejlepsi_pomer) {
          nejlepsi_pomer = pomer;
          nejlepsi_i = i;
        }
      }
    return nejlepsi_i;
  }
  void Start() {
    StartTime();
    konfigurace.Init();
    while (konfigurace.PocetVybranych()<instance.Pocet()) {
      int vec = Najdi();
      konfigurace.Pridej(vec);
      if (konfigurace.Pretizen()) {
        konfigurace.Uber(vec);
        break;
      }
    }
    nejlepsi_konfigurace = konfigurace;
    StopTime();
  }
};

void main(int argc, char *argv[]) {
  if (argc>4) {
    FILE *zadani = fopen(argv[1],"r");
    FILE *reseni = fopen(argv[2],"w");
    FILE *report = fopen(argv[3],"w");
    int algoritmus; sscanf(argv[4],"%d",&algoritmus);
    char line[1024];
    while (fgets(line,1024,zadani)) {
      Instance instance(line);
      if (algoritmus==1) {
        HrubaSila algoritmus(instance);
        algoritmus.Start();
        algoritmus.VypisReseni(reseni);
        algoritmus.VypisReport(report);
      }
      else if (algoritmus==2) {
        HeuristikaCenaVaha algoritmus(instance);
        algoritmus.Start();
        algoritmus.VypisReseni(reseni);
        algoritmus.VypisReport(report);
      }
    }
    fclose(zadani);
    fclose(reseni);
    fclose(report);
  }
}
